@@ -1,3 +1,6 @@
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
 #include <string.h>
 #include <libgnome/gnome-program.h>
 #include <libgnome/gnome-macros.h>
@@ -8,6 +11,8 @@
 #include "eog-collection-item.h"
 #include "eog-image.h"
 #include "eog-canvas-pixbuf.h"
+#include "eog-job-manager.h"
+#include "eog-thumbnail.h"
 
 struct _EogCollectionItemPrivate {
 	EogImage *image;
@@ -34,17 +39,14 @@ static char* get_item_image_caption (GnomeCanvasItem *item, EogImage *image);
 static char* ensure_max_string_width (gchar *str, PangoLayout *layout, int max_width);
 
 
-GNOME_CLASS_BOILERPLATE (EogCollectionItem, eog_collection_item,
-			 GnomeCanvasGroup, GNOME_TYPE_CANVAS_GROUP);
+G_DEFINE_TYPE (EogCollectionItem, eog_collection_item, GNOME_TYPE_CANVAS_GROUP)
 
 static void
 eog_collection_item_class_init (EogCollectionItemClass *klass)
 {
-        GObjectClass *gobject_class;
 	GtkObjectClass *object_class;
 	GnomeCanvasItemClass *item_class;
 
-        gobject_class = (GObjectClass *) klass;
 	object_class = (GtkObjectClass *) klass;
 	item_class = (GnomeCanvasItemClass *) klass;
 
@@ -62,7 +64,7 @@ eog_collection_item_class_init (EogCollectionItemClass *klass)
 }
 
 static void
-eog_collection_item_instance_init (EogCollectionItem *item)
+eog_collection_item_init (EogCollectionItem *item)
 {
 	EogCollectionItemPrivate *priv;
 
@@ -102,7 +104,7 @@ eog_collection_item_destroy (GtkObject *object)
 	    EOG_COLLECTION_ITEM (object)->priv = NULL;
 	}
 
-	GNOME_CALL_PARENT (GTK_OBJECT_CLASS, destroy, (object));
+	GTK_OBJECT_CLASS (eog_collection_item_parent_class)->destroy (object);
 }
 
 static void
@@ -112,7 +114,11 @@ eog_collection_item_update (GnomeCanvasItem *item, double *affine, ArtSVP *clip_
 
 	priv = EOG_COLLECTION_ITEM (item)->priv;
 
-	GNOME_CALL_PARENT (GNOME_CANVAS_ITEM_CLASS, update, (item, affine, clip_path, flags));
+	GNOME_CANVAS_ITEM_CLASS (eog_collection_item_parent_class)->update (
+									item,
+									affine,
+									clip_path,
+									flags);
 
 	if (priv->emit_changed_signal) {
 		priv->emit_changed_signal = FALSE;
@@ -163,6 +169,9 @@ set_pixbuf (EogCollectionItem *item, GdkPixbuf *pixbuf, gboolean view_frame)
 		new_width = image_width * factor;
 		new_height = image_height * factor;
 
+		new_height = MAX (new_height, 1);
+		new_width = MAX (new_width, 1);
+
 		scaled = gnome_thumbnail_scale_down_pixbuf (pixbuf, new_width, new_height);
 
 		image_width = gdk_pixbuf_get_width (scaled);
@@ -170,7 +179,7 @@ set_pixbuf (EogCollectionItem *item, GdkPixbuf *pixbuf, gboolean view_frame)
 	}
 	else {
 		scaled = pixbuf;
-		gdk_pixbuf_ref (scaled);
+		g_object_ref (scaled);
 	}
 
 	g_assert (image_width <= EOG_COLLECTION_ITEM_THUMB_WIDTH);
@@ -198,7 +207,7 @@ set_pixbuf (EogCollectionItem *item, GdkPixbuf *pixbuf, gboolean view_frame)
 		gnome_canvas_item_hide (priv->frame);
 	}
 
-	gdk_pixbuf_unref (scaled);
+	g_object_unref (scaled);
 
 	/* emit size changed signal if pixbuf size changed */
 	if (image_width != old_width || image_height != old_height) {
@@ -226,7 +235,7 @@ get_busy_pixbuf (void)
 		g_free (file);
 	}
 
-	gdk_pixbuf_ref (busy);
+	g_object_ref (busy);
 
 	return busy;
 }
@@ -245,7 +254,7 @@ get_failed_pixbuf (void)
 		g_free (path);
 	}
 
-	gdk_pixbuf_ref (failed);
+	g_object_ref (failed);
 
 	return failed;
 }
@@ -264,32 +273,63 @@ get_stipple_bitmap (void)
 }
 
 static void
-thumbnail_finished_cb (EogImage *image, gpointer data)
+job_thumb_finished (EogJob *job, gpointer data, GError *error)
 {
 	EogCollectionItemPrivate *priv;
+	GdkPixbuf *pixbuf = NULL;
+	gboolean show_frame = FALSE;
+	
+	priv = EOG_COLLECTION_ITEM (data)->priv;
+
+	if (!priv)
+		return;
+		
+	if (eog_job_get_success (job)) {
+		/* success */
+		pixbuf = eog_image_get_pixbuf_thumbnail (priv->image);
+		show_frame = TRUE;
+	}
+	
+	if (pixbuf == NULL) {
+		pixbuf = get_failed_pixbuf ();
+	}
+	
+	g_assert (pixbuf != NULL);
+	
+	set_pixbuf (EOG_COLLECTION_ITEM (data), pixbuf, show_frame);
+	g_object_unref (pixbuf);
+}
+
+static void 
+job_thumb_create (EogJob *job, gpointer data, GError **error)
+{
+	EogCollectionItemPrivate *priv;
+	EogImage *image;
 	GdkPixbuf *pixbuf;
+	GnomeVFSURI *uri;
 
 	priv = EOG_COLLECTION_ITEM (data)->priv;
 
-	pixbuf = eog_image_get_pixbuf_thumbnail (priv->image);
+	image = priv->image; 
+	if (!EOG_IS_IMAGE (image))
+		return;
 
-	set_pixbuf (EOG_COLLECTION_ITEM (data), pixbuf, TRUE);
+	uri = eog_image_get_uri (image);
+	if (uri == NULL)
+		return;
+		
+	pixbuf = eog_thumbnail_load (uri, job, error);
 
-	gdk_pixbuf_unref (pixbuf);
+	if (pixbuf == NULL) 
+		g_assert (*error != NULL);
+	else {
+		eog_image_set_thumbnail (image, pixbuf);
+		g_object_unref (pixbuf);
+	}
+	
+	gnome_vfs_uri_unref (uri);
 }
 
-static void
-thumbnail_failed_cb (EogImage *image, gpointer data)
-{
-	GdkPixbuf *pixbuf;
-
-	pixbuf = get_failed_pixbuf ();
-
-	set_pixbuf (EOG_COLLECTION_ITEM (data), pixbuf, FALSE);
-
-	gdk_pixbuf_unref (pixbuf);
-}	
-
 static void
 image_changed_cb (EogImage *image, gpointer data)
 {
@@ -303,7 +343,7 @@ image_changed_cb (EogImage *image, gpointer data)
 	/* update thumnbail */
 	pixbuf = eog_image_get_pixbuf_thumbnail (priv->image);
 	set_pixbuf (EOG_COLLECTION_ITEM (data), pixbuf, TRUE);
-	gdk_pixbuf_unref (pixbuf);
+	g_object_unref (pixbuf);
 
 	/* update caption */
 	caption = get_item_image_caption (GNOME_CANVAS_ITEM (data), image);
@@ -328,6 +368,7 @@ get_item_image_caption (GnomeCanvasItem *item, EogImage *image)
 	PangoLayout *layout;
 	char *basic_caption;
 	char *caption;
+	char *unescaped;
 
 	g_return_val_if_fail (EOG_IS_IMAGE (image), NULL);
 	g_return_val_if_fail (GNOME_IS_CANVAS_ITEM (item), NULL);
@@ -339,16 +380,18 @@ get_item_image_caption (GnomeCanvasItem *item, EogImage *image)
 	g_assert (layout != NULL);
 
 	/* add line breaks */
-	caption = ensure_max_string_width (basic_caption, layout, EOG_COLLECTION_ITEM_THUMB_WIDTH);
-
+	unescaped = ensure_max_string_width (basic_caption, layout, EOG_COLLECTION_ITEM_THUMB_WIDTH);
+	caption = g_markup_escape_text (unescaped, strlen (unescaped));
+	g_free (unescaped);
+	
 	/* set bold caption to indicate image modification */
 	if (eog_image_is_modified (image)) {
-		char *tmp;
-
-		tmp = g_strdup_printf("<b>%s</b>", caption);
+		char *marked;
+		
+		marked = g_strdup_printf("<b>%s</b>", caption);
 
 		g_free (caption);
-		caption = tmp;
+		caption = marked;
 	}
 
 	g_object_unref (layout);
@@ -429,12 +472,15 @@ ensure_max_string_width (gchar *str, PangoLayout *layout, int max_width)
 	else {
 		char *tail;
 		char *rest;
+		char *tmp;
 
 		/* split remaining rest of string */
 		tail = g_utf8_offset_to_pointer (str, i+1);
 		rest = ensure_max_string_width (tail, layout, max_width);
+		tmp = g_strndup (str, (tail - str));
 
-		result = g_strconcat (g_strndup (str, (tail - str)), "\n", rest, NULL);
+		result = g_strconcat ((tmp), "\n", rest, NULL);
+		g_free (tmp);
 		g_free (rest);
 	}
 
@@ -474,7 +520,7 @@ eog_collection_item_construct (EogCollectionItem *item, EogImage *image)
 						    "justification", GTK_JUSTIFY_CENTER,
 						    "x", (double) caption_x,
 						    "y", (double) caption_y,
-						    "fill_color", style->text[GTK_STATE_NORMAL],
+						    "fill_color_gdk", &style->text[GTK_STATE_NORMAL],
 						    NULL);
 	g_free (caption);
 	     
@@ -507,13 +553,11 @@ eog_collection_item_construct (EogCollectionItem *item, EogImage *image)
 					     NULL);
 	gnome_canvas_item_hide (priv->frame);
 
-	g_signal_connect (image, "thumbnail_failed", G_CALLBACK (thumbnail_failed_cb), item);
-	g_signal_connect (image, "thumbnail_finished", G_CALLBACK (thumbnail_finished_cb), item);
 	g_signal_connect (image, "image_changed", G_CALLBACK (image_changed_cb), item);
 
 	pixbuf = get_busy_pixbuf ();
 	set_pixbuf (item, pixbuf, FALSE);
-	gdk_pixbuf_unref (pixbuf);
+	g_object_unref (pixbuf);
 
 	gnome_canvas_item_request_update (GNOME_CANVAS_ITEM (item));
 }
@@ -568,13 +612,30 @@ void
 eog_collection_item_load (EogCollectionItem *item)
 {
 	EogCollectionItemPrivate *priv;
+	EogJob *job;
+	GdkPixbuf *thumb = NULL;
 
 	g_return_if_fail (EOG_IS_COLLECTION_ITEM (item));
 
 	priv = item->priv;
 
-	if (priv->image != NULL && eog_image_load_thumbnail (priv->image)) {
-		thumbnail_finished_cb (priv->image, item);
+	if (priv->image == NULL) 
+		return;
+
+	thumb = eog_image_get_pixbuf_thumbnail (priv->image);
+	if (thumb == NULL) {
+		/* try to create thumbnail */
+		job = eog_job_new (G_OBJECT (item),
+				   job_thumb_create,
+				   job_thumb_finished,
+				   NULL, /* cancel */
+				   NULL); /* progress */
+		eog_job_manager_add (job);
+		g_object_unref (G_OBJECT (job));
+	}
+	else {
+		set_pixbuf (EOG_COLLECTION_ITEM (item), thumb, TRUE);
+		g_object_unref (thumb);
 	}
 }
 
